/// @custom:security isDelegatecallProxy
contract Proxy{

}

/// @custom:security isUpgradeable
/// @custom:version name=version-0
contract V0{

}

/// @custom:security isUpgradeable
/// @custom:version name=version_1
contract V1{

}