// SPDX-License-Identifier: Unlicense
pragma solidity 0.8.16;

import "./CustomTypeContractLevel.sol";

contract CustomTypeContractLevelUsedInContractTest2
{
    function f(CustomTypeContractLevel.CustomType) public 
    {

    }
}