contract A{

    // ëëëëëëëëëëëëë这这这这这这这这这这

    address unused;


    address unused2;


    // ই এই এই এইই এই এই এইই এই এই এই

    address unused3;


    address unused4;

    // 这这这这这这这这这这这这这这这这这这这这这这这这这这这这这这这这这这这这这这这这这这这这
    address used;


}
