from slither.tools.mutator.mutators.MVIV import MVIV  # severity low
from slither.tools.mutator.mutators.MVIE import MVIE  # severity low
from slither.tools.mutator.mutators.LOR import LOR  # severity medium
from slither.tools.mutator.mutators.UOR import UOR  # severity medium
from slither.tools.mutator.mutators.SBR import SBR  # severity medium
from slither.tools.mutator.mutators.AOR import AOR  # severity medium
from slither.tools.mutator.mutators.BOR import BOR  # severity medium
from slither.tools.mutator.mutators.ASOR import ASOR  # severity medium
from slither.tools.mutator.mutators.MWA import MWA  # severity medium
from slither.tools.mutator.mutators.LIR import LIR  # severity medium
from slither.tools.mutator.mutators.FHR import FHR  # severity medium
from slither.tools.mutator.mutators.MIA import MIA  # severity medium
from slither.tools.mutator.mutators.ROR import ROR  # severity medium
from slither.tools.mutator.mutators.RR import RR  # severity high
from slither.tools.mutator.mutators.CR import CR  # severity high
