contract C {
    function f() public {
        true;
        false;

        123;
        123.456;
        0xabc;

        1 wei;
        1 szabo;
        1 finney;
        1 ether;
        1 seconds;
        1 minutes;
        1 hours;
        1 days;
        1 weeks;
        1 years;

        hex"abcd";

        "abc";
        'def';
    }
}