

contract MinMax {
    uint a = type(uint).max;
    uint b = type(uint).min;

    int constant c = type(int).max;
    int constant d = type(int).min;

    uint8 immutable i =type(uint8).max;
    uint8 immutable j = type(uint8).min;

}
