//pragma solidity ^0.4.24;

// slither-disable-next-line all
contract naming {

    // slither-disable-next-line naming-convention
    struct test {
        uint a;
    }


    // slither-disable-next-line pragma,all
    struct test2 {
        uint a;
    }

    // slither-disable-next-line wrong-naming-convention
    struct test3 {
        uint a;
    }
 
}
