pragma solidity ^0.8.2;

import "./src/ContractV1.sol";
import "./src/ContractV2.sol";
import "./src/InheritedStorageProxy.sol";
import "./src/ERC1967Proxy.sol";
import "./src/TransparentUpgradeableProxy.sol";
