// SPDX-License-Identifier: Unlicense
pragma solidity 0.8.16;

// Top level UsingForDirective not supported yet in Slither. 
// The following code may be uncommented when the support is added.
/*
import "./Library.sol";

using L for uint;
*/

// dummy contract, so that "No contract were found ..." message is not being thrown by Slither
contract Dummy
{

}