bugs_by_version = {
    "0.1.0": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ExpExponentCleanup",
        "ZeroFunctionSelector",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
        "CleanBytesHigherOrderBits",
        "ArrayAccessCleanHigherOrderBits",
        "AncientCompiler",
    ],
    "0.1.1": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ExpExponentCleanup",
        "ZeroFunctionSelector",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
        "CleanBytesHigherOrderBits",
        "ArrayAccessCleanHigherOrderBits",
        "AncientCompiler",
    ],
    "0.1.2": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ExpExponentCleanup",
        "ZeroFunctionSelector",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
        "CleanBytesHigherOrderBits",
        "ArrayAccessCleanHigherOrderBits",
        "AncientCompiler",
    ],
    "0.1.3": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ExpExponentCleanup",
        "ZeroFunctionSelector",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
        "CleanBytesHigherOrderBits",
        "ArrayAccessCleanHigherOrderBits",
        "AncientCompiler",
    ],
    "0.1.4": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
        "CleanBytesHigherOrderBits",
        "ArrayAccessCleanHigherOrderBits",
        "AncientCompiler",
    ],
    "0.1.5": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
        "CleanBytesHigherOrderBits",
        "ArrayAccessCleanHigherOrderBits",
        "AncientCompiler",
    ],
    "0.1.6": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
        "CleanBytesHigherOrderBits",
        "ArrayAccessCleanHigherOrderBits",
        "AncientCompiler",
    ],
    "0.1.7": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
        "CleanBytesHigherOrderBits",
        "ArrayAccessCleanHigherOrderBits",
        "AncientCompiler",
    ],
    "0.2.0": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
        "CleanBytesHigherOrderBits",
        "ArrayAccessCleanHigherOrderBits",
        "AncientCompiler",
    ],
    "0.2.1": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
        "CleanBytesHigherOrderBits",
        "ArrayAccessCleanHigherOrderBits",
        "AncientCompiler",
    ],
    "0.2.2": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
        "CleanBytesHigherOrderBits",
        "ArrayAccessCleanHigherOrderBits",
        "AncientCompiler",
    ],
    "0.3.0": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
        "CleanBytesHigherOrderBits",
        "ArrayAccessCleanHigherOrderBits",
    ],
    "0.3.1": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
        "CleanBytesHigherOrderBits",
    ],
    "0.3.2": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
        "CleanBytesHigherOrderBits",
    ],
    "0.3.3": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
    ],
    "0.3.4": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
    ],
    "0.3.5": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
        "DynamicAllocationInfiniteLoop",
        "OptimizerClearStateOnCodePathJoin",
    ],
    "0.3.6": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
        "SendFailsForZeroEther",
    ],
    "0.4.0": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
        "LibrariesNotCallableFromPayableFunctions",
    ],
    "0.4.1": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
        "LibrariesNotCallableFromPayableFunctions",
    ],
    "0.4.10": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
    ],
    "0.4.11": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
    ],
    "0.4.12": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
    ],
    "0.4.13": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
    ],
    "0.4.14": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
    ],
    "0.4.15": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
    ],
    "0.4.16": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
    ],
    "0.4.17": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "EventStructWrongData",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
    ],
    "0.4.18": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "EventStructWrongData",
        "NestedArrayFunctionCallDecoder",
    ],
    "0.4.19": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ABIEncoderV2PackedStorage_0.4.x",
        "ExpExponentCleanup",
        "EventStructWrongData",
        "NestedArrayFunctionCallDecoder",
    ],
    "0.4.2": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
        "OptimizerStaleKnowledgeAboutSHA3",
    ],
    "0.4.20": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ABIEncoderV2PackedStorage_0.4.x",
        "ExpExponentCleanup",
        "EventStructWrongData",
        "NestedArrayFunctionCallDecoder",
    ],
    "0.4.21": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ABIEncoderV2PackedStorage_0.4.x",
        "ExpExponentCleanup",
        "EventStructWrongData",
        "NestedArrayFunctionCallDecoder",
    ],
    "0.4.22": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ABIEncoderV2PackedStorage_0.4.x",
        "ExpExponentCleanup",
        "EventStructWrongData",
        "OneOfTwoConstructorsSkipped",
    ],
    "0.4.23": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ABIEncoderV2PackedStorage_0.4.x",
        "ExpExponentCleanup",
        "EventStructWrongData",
    ],
    "0.4.24": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ABIEncoderV2PackedStorage_0.4.x",
        "ExpExponentCleanup",
        "EventStructWrongData",
    ],
    "0.4.25": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ABIEncoderV2PackedStorage_0.4.x",
    ],
    "0.4.26": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
    ],
    "0.4.3": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "HighOrderByteCleanStorage",
    ],
    "0.4.4": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
    ],
    "0.4.5": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
        "OptimizerStateKnowledgeNotResetForJumpdest",
    ],
    "0.4.6": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
        "IdentityPrecompileReturnIgnored",
    ],
    "0.4.7": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
    ],
    "0.4.8": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
    ],
    "0.4.9": [
        "DirtyBytesArrayToStorage",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "UninitializedFunctionPointerInConstructor_0.4.x",
        "IncorrectEventSignatureInLibraries_0.4.x",
        "ExpExponentCleanup",
        "NestedArrayFunctionCallDecoder",
        "ZeroFunctionSelector",
        "DelegateCallReturnValue",
        "ECRecoverMalformedInput",
        "SkipEmptyStringLiteral",
        "ConstantOptimizerSubtraction",
    ],
    "0.5.0": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor",
        "IncorrectEventSignatureInLibraries",
        "ABIEncoderV2PackedStorage",
    ],
    "0.5.1": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor",
        "IncorrectEventSignatureInLibraries",
        "ABIEncoderV2PackedStorage",
    ],
    "0.5.10": [
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "YulOptimizerRedundantAssignmentBreakContinue0.5",
        "ABIEncoderV2CalldataStructsWithStaticallySizedAndDynamicallyEncodedMembers",
    ],
    "0.5.11": [
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "YulOptimizerRedundantAssignmentBreakContinue0.5",
    ],
    "0.5.12": [
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "YulOptimizerRedundantAssignmentBreakContinue0.5",
    ],
    "0.5.13": [
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "YulOptimizerRedundantAssignmentBreakContinue0.5",
    ],
    "0.5.14": [
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "MissingEscapingInFormatting",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "YulOptimizerRedundantAssignmentBreakContinue0.5",
        "ABIEncoderV2LoopYulOptimizer",
    ],
    "0.5.15": [
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "MissingEscapingInFormatting",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "YulOptimizerRedundantAssignmentBreakContinue0.5",
    ],
    "0.5.16": [
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "MissingEscapingInFormatting",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
    ],
    "0.5.17": [
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "MissingEscapingInFormatting",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
    ],
    "0.5.2": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor",
        "IncorrectEventSignatureInLibraries",
        "ABIEncoderV2PackedStorage",
    ],
    "0.5.3": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor",
        "IncorrectEventSignatureInLibraries",
        "ABIEncoderV2PackedStorage",
    ],
    "0.5.4": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor",
        "IncorrectEventSignatureInLibraries",
        "ABIEncoderV2PackedStorage",
    ],
    "0.5.5": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor",
        "IncorrectEventSignatureInLibraries",
        "ABIEncoderV2PackedStorage",
        "IncorrectByteInstructionOptimization",
        "DoubleShiftSizeOverflow",
    ],
    "0.5.6": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "ABIEncoderV2CalldataStructsWithStaticallySizedAndDynamicallyEncodedMembers",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor",
        "IncorrectEventSignatureInLibraries",
        "ABIEncoderV2PackedStorage",
        "IncorrectByteInstructionOptimization",
    ],
    "0.5.7": [
        "DirtyBytesArrayToStorage",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "ABIEncoderV2CalldataStructsWithStaticallySizedAndDynamicallyEncodedMembers",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
        "UninitializedFunctionPointerInConstructor",
        "IncorrectEventSignatureInLibraries",
    ],
    "0.5.8": [
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "YulOptimizerRedundantAssignmentBreakContinue0.5",
        "ABIEncoderV2CalldataStructsWithStaticallySizedAndDynamicallyEncodedMembers",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
        "DynamicConstructorArgumentsClippedABIV2",
    ],
    "0.5.9": [
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "privateCanBeOverridden",
        "YulOptimizerRedundantAssignmentBreakContinue0.5",
        "ABIEncoderV2CalldataStructsWithStaticallySizedAndDynamicallyEncodedMembers",
        "SignedArrayStorageCopy",
        "ABIEncoderV2StorageArrayWithMultiSlotElement",
    ],
    "0.6.0": [
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "MissingEscapingInFormatting",
        "ArraySliceDynamicallyEncodedBaseType",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
        "YulOptimizerRedundantAssignmentBreakContinue",
    ],
    "0.6.1": [
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "MissingEscapingInFormatting",
        "ArraySliceDynamicallyEncodedBaseType",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
    ],
    "0.6.10": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
    ],
    "0.6.11": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
    ],
    "0.6.12": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
    ],
    "0.6.2": [
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "MissingEscapingInFormatting",
        "ArraySliceDynamicallyEncodedBaseType",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
    ],
    "0.6.3": [
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "MissingEscapingInFormatting",
        "ArraySliceDynamicallyEncodedBaseType",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
    ],
    "0.6.4": [
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "MissingEscapingInFormatting",
        "ArraySliceDynamicallyEncodedBaseType",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
        "MemoryArrayCreationOverflow",
    ],
    "0.6.5": [
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "MissingEscapingInFormatting",
        "ArraySliceDynamicallyEncodedBaseType",
        "ImplicitConstructorCallvalueCheck",
        "TupleAssignmentMultiStackSlotComponents",
    ],
    "0.6.6": [
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "MissingEscapingInFormatting",
        "ArraySliceDynamicallyEncodedBaseType",
        "ImplicitConstructorCallvalueCheck",
    ],
    "0.6.7": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "MissingEscapingInFormatting",
        "ArraySliceDynamicallyEncodedBaseType",
        "ImplicitConstructorCallvalueCheck",
    ],
    "0.6.8": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
    ],
    "0.6.9": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "UsingForCalldata",
    ],
    "0.7.0": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
    ],
    "0.7.1": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
        "FreeFunctionRedefinition",
    ],
    "0.7.2": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
        "DynamicArrayCleanup",
    ],
    "0.7.3": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
        "EmptyByteArrayCopy",
    ],
    "0.7.4": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
    ],
    "0.7.5": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
    ],
    "0.7.6": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
    ],
    "0.8.0": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
    ],
    "0.8.1": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
    ],
    "0.8.10": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
    ],
    "0.8.11": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "AbiEncodeCallLiteralAsFixedBytesBug",
    ],
    "0.8.12": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "AbiEncodeCallLiteralAsFixedBytesBug",
    ],
    "0.8.13": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "StorageWriteRemovalBeforeConditionalTermination",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "InlineAssemblyMemorySideEffects",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
    ],
    "0.8.14": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "StorageWriteRemovalBeforeConditionalTermination",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "InlineAssemblyMemorySideEffects",
    ],
    "0.8.15": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "StorageWriteRemovalBeforeConditionalTermination",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
    ],
    "0.8.16": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "StorageWriteRemovalBeforeConditionalTermination",
    ],
    "0.8.17": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
    ],
    "0.8.18": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
    ],
    "0.8.19": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
    ],
    "0.8.2": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
        "KeccakCaching",
    ],
    "0.8.20": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
    ],
    "0.8.21": ["VerbatimInvalidDeduplication"],
    "0.8.22": ["VerbatimInvalidDeduplication"],
    "0.8.23": [],
    "0.8.24": [],
    "0.8.25": [],
    "0.8.3": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
        "ABIDecodeTwoDimensionalArrayMemory",
    ],
    "0.8.4": [
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
    ],
    "0.8.5": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
    ],
    "0.8.6": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
    ],
    "0.8.7": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "SignedImmutables",
    ],
    "0.8.8": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
        "UserDefinedValueTypesBug",
        "SignedImmutables",
    ],
    "0.8.9": [
        "VerbatimInvalidDeduplication",
        "FullInlinerNonExpressionSplitArgumentEvaluationOrder",
        "MissingSideEffectsOnSelectorAccess",
        "AbiReencodingHeadOverflowWithStaticArrayCleanup",
        "DirtyBytesArrayToStorage",
        "DataLocationChangeInInternalOverride",
        "NestedCalldataArrayAbiReencodingSizeValidation",
    ],
}
