"""
This module import all slither exceptions
"""

from slither.slithir.exceptions import SlithIRError as SlithIRError
from slither.solc_parsing.exceptions import ParsingError as ParsingError
from slither.solc_parsing.exceptions import VariableNotFound as VariableNotFound
from slither.core.exceptions import SlitherCoreError as SlitherCoreError
from slither.exceptions import SlitherException as SlitherException
